<?php

/*
Plugin Name: Twitter Digest
Plugin URI: http://whalespine.org/projects/wordpress/twitter-digest
Description:  Create a daily or weekly post on your blog containing the tweets from your <a href="http://twitter.com">Twitter</a> account.

Version: 2.5
Author: Tim Beck, with input from Paul Wlodarczyk 
Author URI: http://whalespine.org
*/

// Copyright (c) 2009 - 2010 Tim Beck and Paul Wlodarczyk. All rights reserved.
//
// Released under the GPL license
// http://www.opensource.org/licenses/gpl-license.php
//
// This is an add-on for WordPress
// http://wordpress.org/
//
// **********************************************************************
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
// **********************************************************************
//
//=======================
// Defines

define('WS_FETCH_SPACE', 30); //X seconds between fetches 
define('WS_TWITTER_LIMIT', 200); //the most tweets you can pull from the Twitter REST API  
define('WS_API_USER_TIMELINE', 'http://api.twitter.com/statuses/user_timeline.json');
define('WS_STATUS_URL', 'http://twitter.com/###USERNAME###/statuses/###STATUS###');
define('WS_PROFILE_URL', 'http://twitter.com/###USERNAME###');
define('WS_HASHTAG_URL', 'http://search.twitter.com/search?q=###HASHTAG###');
define('WS_VERSION', '2.5');

//========================
// Digest cron activation

register_activation_hook(__FILE__, 'ws_activate_digest');
register_deactivation_hook(__FILE__, 'ws_deactivate_digest');
add_action('ws_td_digest_event', 'ws_ping_twitter');

function ws_activate_digest() {
  wp_schedule_event(time(), 'hourly', 'ws_td_digest_event');
}

function ws_deactivate_digest() {
  wp_clear_scheduled_hook('ws_td_digest_event');
}

// Check Twitter to see if there are any new tweets that need to be published.
// Return values:
// null -> all good
// 1 -> Error w/ twitter
// 2 -> Too soon
// 3 -> Missing username/password
// 4 -> Post created and published
// 5 -> No tweets to post
// 6 -> Post created but not published until later
// 7 -> Wrong day of week

function ws_ping_twitter() {

  // Do we have a username and password?
  if (!get_option("ws_td_username")) {
    return 3;
  }

  // We don't need the password anymore, so delete it.
  if (get_option("ws_td_password")) {
     delete_option("ws_td_password");
  }
  
 
  // Has there been enough time since the last check? This avoids a race
  // condition which would produce a duplicate post.

  $last = get_option('ws_td_last_check');
  if (time() - $last < WS_FETCH_SPACE) {
    return 2;
  }
  update_option("ws_td_last_check", time());

  // Is this a daily or weekly post?

  // Need to know today's local midnight to work out start/end times
  $midnight = strtotime(date('Y-m-d, 00:00:00', time() + ws_gmtOffset()));
  
  $pub_day = get_option('ws_td_pub_day');
  $pub_freq = get_option('ws_td_freq');
  $current_day_of_week = date("w", $midnight);

  // Test if it's weekly, then if so is today the right day of the week. If
  // not, return; pub_freq=0 is daily
  if ($pub_freq == 1) {
    if ($current_day_of_week != $pub_day) {
      return 7;
    } else {
      // 7 days ago at midnight
      $startDate = strtotime('-7 days', $midnight);
    }
  } else {
    // yesterday at midnight
    $startDate = strtotime('-1 days', $midnight);
  }

  // yesterday at 11:59:59
  $endDate = strtotime('-1 second', $midnight);


  $startDateStr = date(ws_getDateFormat(), $startDate);
  $endDateStr = date(ws_getDateFormat(), $endDate);
  
  // Get the last tweet id from the last post
  $lastTweet = get_option('ws_td_last_tweet');
  if (!$lastTweet) {
    $lastTweet = 0;
    update_option('ws_td_last_tweet', $lastTweet);
  }

  // Get the tweets
  $numtweets = get_option('ws_td_num_tweets');
  // range check. Twitter limit is to 200
  if ($numtweets > WS_TWITTER_LIMIT) { 
    $numtweets = WS_TWITTER_LIMIT;
  }
  
  if ($numtweets == 0) {
    $numtweets = 20;  //default is 20 with no count argument, so fetch 20 if NULL
  }

  // get the last N tweets, since the last tweet
  $url = WS_API_USER_TIMELINE . "?screen_name=".get_option('ws_td_username') ."&count=". $numtweets ;
  if ($lastTweet) {
    $url .= "&since_id=" . $lastTweet; 
  }

  
  // Fetch the tweets
  $tweets = ws_fetch_tweets($url);

  // Go through the array and process any tweets from the desired time period 
  $tweet_content = array();
  if (count($tweets) >= 0) { 
    // process the tweets

    foreach ($tweets as $tw_data) {
      // Convert the time to local
      $tw_data->created_at = strtotime($tw_data->created_at) + ws_gmtOffset();


      // Was this tweet added in the time period of interest? 
      if ($tw_data->created_at < $startDate) {
        continue;
      }
      if ($tw_data->created_at > $endDate ) {
        continue;
      }

      // Are we dropping replies?
      if (get_option('ws_td_drop_replies') && preg_match('/^@.*/', $tw_data->text)) {
        continue;
      }
      
      // All good, so format and add to the content
      array_push($tweet_content, ws_format_tweet($tw_data)."\n\n");

      // Remember the first tweet in the list, which is actually the most
      // recently tweeted.
      if (!$newLastTweet) {
        $newLastTweet = $tw_data->id;
      }
    }

    
    // To avoid race conditions, check to make sure lastTweet hasn't changed
    if ($lastTweet == get_option('ws_td_last_tweet')) {

      $count = count($tweet_content);

      if ($count != 0 && $count >= get_option('ws_td_min_tweets')) {

        // Before we push the content, we need to put it in the right order
        if (get_option('ws_td_chrono') == 1) {
          $tweet_content = array_reverse($tweet_content);
        }   
        
        $title = get_option('ws_td_title');
        if (!$title) {
          $title = "Tweets for %startdate";
        }

        // We look for %date to be backwards compatible
        $title = preg_replace("/%date/", $startDateStr, $title);        
        $title = preg_replace("/%startdate/", $startDateStr, $title);
        $title = preg_replace("/%enddate/", $endDateStr, $title);        

        $excerpt = get_option('ws_td_excerpt');
        $excerpt = preg_replace("/%date/", $startDateStr, $excerpt);
        $excerpt = preg_replace("/%startdate/", $startDateStr, $excerpt);
        $excerpt = preg_replace("/%enddate/", $endDateStr, $excerpt);        
        
        $postcontent = "<ul class=\"ws_tweet_list\">\n\n";
        $postcontent .= join("\n\n", $tweet_content);
        $postcontent .= "\n</ul>";

        $pubtime = get_option('ws_td_pub_time');
        if ($pubtime) {
          ws_create_post($postcontent, $title, $pubtime, $excerpt);
          $retval = 6;
        } else {
          ws_create_post($postcontent,$title, null, $excerpt);
          $retval = 4;
        }
        
        // Update the last tweet id
        update_option('ws_td_last_tweet', $newLastTweet);

        return $retval;
        
      }
      else {
        return 5;
      }
    }
    
  }
  else {
    return 5;
  } 
   
}

// This function creates the actual post and schedules it for publishing time
// at $pubtime
function ws_create_post($content, $title, $pubtime, $excerpt) {

  global $wpdb;

  // Are we doing this now or later?
  if ($pubtime) {
    $time = date('Y-m-d ').$pubtime.":00";
    $now = date('Y-m-d H:i:s');
    if ($time > $now) {
      $status = 'future';
    } else {
      $status = 'publish';
    }
  } else {
    $status = 'publish';
    $time = date('Y-m-d, H:i:s');
  }

  
  // Create the post
  $post_data = array(
                     'post_content' => $wpdb->escape($content),
                     'post_title' => $wpdb->escape($title),
                     'post_date' => $time,
                     'post_category' => array(get_option('ws_td_category')),
                     'post_status' => $status,
                     'post_author' => $wpdb->escape(get_option('ws_td_author')),
                     'post_excerpt' => $wpdb->escape($excerpt)
                     );
  
  // Insert post
  $post_id = wp_insert_post($post_data);
  add_post_meta($post_id, 'ws_tweeted', '1', true);
  
  // Tag it
  wp_set_post_tags($post_id, get_option('ws_td_post_tags'));
  
}

// Returns an html formatted $tweet.  This is almost directly borrowed from Twitter Tools
function ws_format_tweet($tweet) {
  
  $output = '<li class="ws_tweet">';
  $output .= ws_make_clickable(wp_specialchars($tweet->text));
  if (!empty($tweet->in_reply_to_screen_name)
      && (!empty($tweet->in_reply_to_status_id))) {
    $output .= 	' <a href="'.ws_status_url($tweet->in_reply_to_screen_name, $tweet->in_reply_to_status_id).'">'.sprintf(__('in reply to %s', 'twitter-digest'), $tweet->in_reply_to_screen_name).'</a>';
  }

  // Show the date/time if the options are selected
  $showTime = get_option('ws_td_showtime');
  if (!$showTime) { $showTime = 0; }
  // Show the date if the option is selectd
  $showDate = get_option('ws_td_showdate');
  if (!$showDate) { $showDate = 0; }
  
  $time_display = '';
  if ($showTime == 1) {
    $time_display = gmdate('H:i:s', $tweet->created_at);

    // Add the comma if we are going to show the date as well
    if ($showDate == 1) {
      $time_display .= ", ";
    }
  }

  // Add the date
  if ($showDate == 1) {
    $time_display .= gmdate('Y-m-d', $tweet->created_at);
  }

  // Use a small arrow for the status links if time and date are off
  if ($showDate != 1 && $showTime != 1) {
    $time_display = "->";
  }

  // Add the status link 
  $username = get_option('ws_td_username'); 
  $output .= ' <a class="ws_tweet_time" href="'.ws_status_url($username, $tweet->id).'">'.$time_display.'</a>';

      
  $output .= '</li>';
  return $output;
  
}

// Most of the following formatting functions are borrowed from Twitter Tool
function ws_make_clickable($tweet) {
  
  //	$tweet = preg_replace('/\@([a-zA-Z0-9_]{1,15}) /','@<a
  //	href="http://twitter.com/\\1">\\1</a> ', $tweet);
  $tweet = preg_replace_callback(
                                 '/((?:^|\s+|[\"\'\[\(]))\@([\w]{1,15})/',
                                 create_function(
                                                 '$matches',
                                                 'return $matches[1].ws_profile_link($matches[2]);'
                                                 ),
                                 $tweet
                                 );
  $tweet = preg_replace_callback(
                                 '/((?:^|\s+|[\"\'\[\(]))\#([\w]{1,15})/',
                                 create_function(
                                                 '$matches',
                                                 'return $matches[1].ws_hashtag_link($matches[2]);'
                                                 ),
                                 $tweet
                                 );

  // make_clickable doesn't handle (url) very well, so we'll chuck some spaces in
  $tweet = preg_replace("/\((http.*)\)/", "( $1 )", $tweet);

  return make_clickable($tweet);
}

function ws_status_url($username, $status) {
  return str_replace(
                     array('###USERNAME###', '###STATUS###'),
                     array($username, $status),
                     WS_STATUS_URL);
}

function ws_profile_link($username, $prefix = '@') {
  return '<a href="'.ws_profile_url($username).'">'.$prefix.$username.'</a>';
}

function ws_hashtag_link($hashtag, $prefix = '#') {
  return '<a href="'.ws_hashtag_url($hashtag).'">'.$prefix.htmlspecialchars($hashtag).'</a>';
}

function ws_profile_url($username) {
  return str_replace('###USERNAME###', $username, WS_PROFILE_URL);
}

function ws_hashtag_url($hashtag) {
  $hashtag = urlencode('#'.$hashtag);
  return str_replace('###HASHTAG###', $hashtag, WS_HASHTAG_URL);
}

// Based on Alex King's implementation for the Twitter Tools plugin
function ws_fetch_tweets($url) {
  
  require_once(ABSPATH.WPINC.'/class-snoopy.php');
  $snoop = new Snoopy;
  $snoop->agent = 'Whalespine';
  $snoop->fetch($url);
  
  if (!strpos($snoop->response_code, '200')) {
     update_option('ws_td_error', 'Error retrieving tweets. Make sure your username and password are correct.<br/> Status: '.$snoop->status.' <br/> '.$snoop->results);
     return false;
  }

  // Everything is ok
  update_option('ws_td_error','');

  // To parse the json we use the build in function if we have it.
  // Note that Services_JSON isn't the fastest thing in the world and has
  // known to time out PHP on large json strings
  if (function_exists('json_decode')) {
    $tweets = json_decode($snoop->results);
  } else {
    // Process the result
    $json = new Services_JSON();
    $json->decode($snoop->results);    
  }
  
  return $tweets;
}

// Simply resets the 'ws_td_last_tweet' option
function ws_td_resetDatabase() {
  update_option('ws_td_last_tweet',0);   
}

// The menu hook
function ws_menu_item() {
  if (current_user_can('manage_options')) {
    add_options_page(
                     __('Twitter Digest Options', 'twitter-digest')
                     , __('Twitter Digest', 'twitter-digest')
                     , 10
                     , basename(__FILE__)
                     , 'ws_options_form'
                     );
  }
}
add_action('admin_menu', 'ws_menu_item');


function ws_options_form() {

  // Check here if we are going to do the check now
  global $wpdb;

  // Reset the database if necessary
  if ($_POST["action"] == 'resetdb') {
     ws_td_resetDatabase();
  }

  // Get some variables together
  $ping_message = "(This may take a while if there are tweets to process.)";
  $ping_style="color: black";
  
  // Range check the publish day  
  $pub_day_value = get_option('ws_td_pub_day');
  if ($pub_day_value < 1 || $pub_day_value > 6) { $pub_day_value =  date("w"); }

  // yesterday or last week?
  if (get_option('ws_td_freq') == 0) {
    $periodStr = 'yesterday';
  } else {
    $periodStr = 'last week';
  }
  
  // Ping if necessary, and show the correct message.
  if ($_POST["action"] == "ping") {
    switch(ws_ping_twitter(true)) {
    case 1:
      $pwuser = get_option('ws_td_username');
      
      $ping_message = "Something went wrong with Twitter.  Check for an error message above. Current Twitter username is " . $pwuser. "." ;
      $ping_style = "color: red";
      break;
    case 2:
      $ping_message = "To keep things sane, we're going to wait ".WS_FETCH_SPACE." seconds between pings.";
      $ping_style = "color: red";
      break;
    case 3:
      $ping_message = "Missing Twitter username and/or password";
      $ping_style = "color: red";
      break;
    case 4:
      $ping_message = "Post containing ".$periodStr."'s tweets published.";
      $ping_style ="color: green";
      break;
    case 6: $ping_message = "Post containing ".$periodStr ."'s tweets scheduled.";
      $ping_style = "color: green";
      break;
    case 5: $ping_message = "No new posts found from " . $periodStr;
	  break;
    case 7:
      $daylabel = date("l");
      $pub_day_str = ws_getDoW($pub_day_value);
      $now = date("H:i:s");
      $pub_time_str = get_option('ws_td_pub_time');
      if (!$pub_time_str) { $pub_time_str = '00:00:00'; }
      
      $ping_message = "Today is " . $daylabel . " and the local time is " . $now . " and you are configured to publish on " . $pub_day_str . " at " . $pub_time_str . ".";
     $ping_style = "color: red";
    }
  }
  
  // Get all the options together
  $categories = get_categories('hide_empty=0');
  $cat_options = '';
  foreach ($categories as $category) {
    if ($category->term_id == get_option('ws_td_category')) {
      $selected = 'selected="selected"';
    }
    else {
      $selected = '';
    }
    $cat_options .= "\n\t<option value='$category->term_id' $selected>$category->name</option>";
  }
  
  $authors = get_users_of_blog();
  $author_options = '';
  foreach ($authors as $user) {
    $usero = new WP_User($user->user_id);
    $author = $usero->data;
    // Only list users who are allowed to publish
    if (! $usero->has_cap('publish_posts')) {
      continue;
    }
    if ($author->ID == get_option('ws_td_author')) {
      $selected = 'selected="selected"';
    }
    else {
      $selected = '';
    }
    $author_options .= "\n\t<option value='$author->ID' $selected>$author->user_nicename</option>";
  }
  
  // Drop replies options
  $drop_replies_check = get_option('ws_td_drop_replies') == 1 ? 'checked="true"' : '';
  // Chrono options
  $chrono_check = get_option('ws_td_chrono') == 1 ? 'checked="true"' : '';
  // Show date options
  $showDate_check = get_option('ws_td_showdate') == 1 ? 'checked="true"' : '';
  $showTime_check = get_option('ws_td_showtime') == 1 ? 'checked="true"' : '';
  
  // Publishing frequency options
  if (get_option('ws_td_freq') == 1) {
    $freq_options = '
       <option value="0">'.__('Daily', 'twitter-digest').'</option>
       <option value="1" selected="selected">'.__('Weekly', 'twitter-digest').'</option>
     ';

    $pub_day_display = '';
  }  else {
    $freq_options = '
      <option value="0" selected="selected">'.__('Daily', 'twitter-digest').'</option>
      <option value="1">'.__('Weekly', 'twitter-digest').'</option>
    ';
    $pub_day_display = 'display: none';
  }

  // DoW option
  $pub_day_options = '';
  for ($i = 0; $i < 7; $i++) {
    $pub_day_options .= '<option value="' . $i . '"';
    if ($i == $pub_day_value) {
      $pub_day_options .= ' selected="selected"';
    }
    $pub_day_options .= '>' . ws_getDow($i) . '</option>';
  }
  
  // Default the min tweet value to 1
  $min_tweets_value = get_option('ws_td_min_tweets');
  if ($min_tweets_value < 1) { $min_tweets_value = 1; }

  // Default the number of tweets to 20
  $num_tweets_value = get_option('ws_td_num_tweets');
  if (!$num_tweets_value) { $num_tweets_value =  20; }

  print('
  
  <style>
  
  p.submit {
  padding: 5px;
margin-top: 20px;
  }
  div.ws_td_error {
  margin: 20px 100px 20px 50px;    
  padding: 10px;
  border: 2px solid red;
    background-color: #FFEEEE;
  }
  .option {margin: 10px;}
  .section {
    border-bottom: 1px dashed #888;
    margin-bottom: 10px;
    padding-bottom: 10px;
  }
  fieldset h3 {
    color: #888;
  }
  </style>


  <script text="text/javascript">
    var ws = {
      // Toggle the day of week option
      toggleDoWOption: function(select) {

         if (jQuery(select).val() == 0) {
            jQuery("#dowOptionDiv").hide();
        } else {
            jQuery("#dowOptionDiv").show();
        }
      }
    }
  </script>

  <div class="wrap">
  <h2>'.__('Twitter Digest Options', 'twitter-tools').'</h2>
  
  '.ws_getErrorMessage().'
    <hr/><form id="ws_twittertools" name="ws_twittertools" action="options.php" method="post">
  '.wp_nonce_field('update-options').'
    
    
    <input type="hidden" name="action" value="update" />
    <input type="hidden" name="page_options" value="ws_td_username,ws_td_category,ws_td_author,ws_td_title,ws_td_drop_replies,ws_td_post_tags,ws_td_min_tweets,ws_td_chrono,ws_td_pub_time,ws_td_pub_day,ws_td_freq,ws_td_num_tweets,ws_td_showdate,ws_td_showtime, ws_td_excerpt, ws_td_dateFormat, ws_td_checktest"/>
  
  <fieldset class="options section">
  <h3>Twitter Account Info</h3>
  <div class="option">
  
  <label for="ws_td_username">'.__('Username', 'twitter-digest').': </label>
  <input type="text" size="25" name="ws_td_username" id="ws_td_username" value="'.get_option('ws_td_username').'" autocomplete="off" />
  </div>

  </fieldset>


  <fieldset class="options section">
  <h3>Publish Options</h3>

  <div class="option">
    <label for="ws_td_freq">'.__('Frequency', 'twitter-digest').': </label>
    <select name="ws_td_freq" id="ws_td_freq" onchange="ws.toggleDoWOption(this);">'.$freq_options.'</select>
  </div>


  <div class="option">
  <label for="ws_td_pub_time">'.__('Time of Day:', 'twitter-digest').'</label>
  <input name="ws_td_pub_time" id="ws_td_pub_time" style="text-align:right" size="5" value="'.get_option('ws_td_pub_time').'">
  <span>'._('Specify the time at which the tweet post should be published in 24 hour format e.g. 2:22pm = 14:22 or leave empty for ASAP').'</span>
  </div>
  <div class="option" id="dowOptionDiv" style="'. $pub_day_display.'">
  <label for="ws_td_pub_day">'.__('Day of Week', 'twitter-digest').'</label>
  <select name="ws_td_pub_day" id="ws_td_pub_day">' . $pub_day_options . '</select>

  <span>'._('Specify the day of the week the post should be published in number format (0-6). e.g. Sunday = 0').'</span>
  </div>
  

   </fieldset>

  <fieldset class="options section">
  <h3>Post Options</h3>

  <div class="option">
  <label for="ws_td_title">'.__('Post Title:', 'twitter-digest').'</label>
  <input size="50" name="ws_td_title" id="ws_td_title" value="'.get_option('ws_td_title').'">
  <span>'._('Use %startdate and %enddate to insert the date').'</span>
  </div>

  <div class="option">
  <label for="ws_td_excerpt">'.__('Post Excerpt:', 'twitter-digest').'</label>
  <input size="75" name="ws_td_excerpt" id="ws_td_excerpt" value="'.get_option('ws_td_excerpt').'">
  <span>'._('Use %startdate and %enddate to insert the dates').'</span>
  </div>

  <div class="option">
  <label for="ws_td_dateFormat">'.__('Post Title/Excerpt Date Format:', 'twitter-digest').'</label>
  <input type="text" size="10" name="ws_td_dateFormat" id="ws_td_dateFormat" value="'.ws_getDateFormat().'"/>
  <span>'._('Defaults to Y-m-d').'</span>
  </div>

  <div class="option">
  <label for="ws_td_category">'.__('Post Category:', 'twitter-digest').'</label>
  <select name="ws_td_category" id="ws_td_category">'.$cat_options.'</select>
  </div>
  <div class="option">
  <label for="ws_td_post_tags">'.__('Post Tags:', 'twitter-digest').'</label>
  <input size="50" name="ws_td_post_tags" id="ws_td_post_tags" value="'.get_option('ws_td_post_tags').'">
  <span>'._('Separate multiple tags with commas. Example: tweets, twitter').'</span>
  </div>
  <div class="option">
  <label for="ws_td_author">'.__('Post Author:', 'twitter-digest').'</label>
  <select name="ws_td_author" id="ws_blog_post_author">'.$author_options.'</select>
  </div>

  <div class="option">
  <label for="ws_td_drop_replies">'.__('Exclude @reply tweets from the digest?', 'twitter-digest').'</label>
  <input value="1" type="checkbox" name="ws_td_drop_replies" id="ws_td_drop_replies" '.$drop_replies_check.'/_check>
  </div>
  <div class="option">
  <label for="ws_td_chrono">'.__('Show tweets in chronological order?', 'twitter-digest').'</label>
  <input value="1" type="checkbox" name="ws_td_chrono" id="ws_td_chrono" '.$chrono_check.'/>
  </div>
  <div class="option">
  <label for="ws_td_showdate">'.__('Show tweet dates?', 'twitter-digest').'</label>
  <input value="1" type="checkbox" name="ws_td_showdate" id="ws_td_showdate" '.$showDate_check.'/>
  </div> 
  <div class="option">
  <label for="ws_td_showtime">'.__('Show tweet times?', 'twitter-digest').'</label>
  <input value="1" type="checkbox" name="ws_td_showtime" id="ws_td_showtime" '.$showTime_check.'/>
  </div>  

    
  <div class="option">
  <label for="ws_td_min_tweets">'.__('Minimum number of tweets required for post: ', 'twitter-digest').'</label>
  <input name="ws_td_min_tweets" id="ws_td_min_tweets" size="1" style="text-align: right" value="'.$min_tweets_value.'">
  </div>
  
  <div class="option">
  <label for="ws_td_num_tweets">'.__('Maximum number of tweets to retrieve (Twitter caps at 200): ', 'twitter-digest').'</label>
  <input name="ws_td_num_tweets" id="ws_td_num_tweets" size="3" style="text-align: right" value="'.$num_tweets_value.'">
  </div>
  </fieldset>
  <div class="section">
  <p class="submit">
  <input type="submit" name="submit" value="'.__('Update Options', 'twitter-digest').'" />
  </p>
  </div>

  </form>

   <div class="section"
  <form method="POST">
    <p class="submit" style="margin: 0;">
    <input type="hidden" name="action" value="ping"/>
    <input type="submit" name="submit" value="'.__('Ping Twitter', 'twitter-digest').'" />
    <span style="'.$ping_style.'">'.$ping_message.'</span>
  
  </p>
  </form>
  <form method="POST">
  <p class="submit" style="margin: 0;">
  <input type="hidden" name="action" value="resetdb"/>
  <input type="submit" name="submit" value="'.__('Reset Database', 'twitter-digest').'" />
  <span>Clicking this button resets the Twitter Digest database and may result in duplicate posts.</span>
  </p>
  </form>
</div>
  
</div>
  ');
}


// Get the date option or fall back to the default
function ws_getDateFormat() {
  $format = get_option('ws_td_dateFormat');
  if ($format == '') {
    $format = 'Y-m-d';
  }
  return $format;
}

// Get the string rep of a day given the index. Sunday = 0;
function ws_getDoW($num) {
  switch($num) {
  case 0:
    return "Sunday";
    break;
  case 1:
    return "Monday";
    break;
  case 2:
    return "Tuesday";
    break;
  case 3:
    return "Wednesday";
    break;
  case 4:
    return "Thursday";
    break;
  case 5: 
    return "Friday";
    break;
  case 6: 
    return "Saturday";
  }
  
}

// Returns the gmt offset for Wordpress in seconds
function ws_gmtOffset() {
  return get_option('gmt_offset') * 3600;
}

function ws_gmtTime() {
  return time() - date("Z");
}

function ws_getErrorMessage() {
  $error = get_option('ws_td_error');
  if ($error) {
    return '<div class="ws_td_error">'.$error.'</div>';
  }
}

define('MYPLUGIN_FOLDER', dirname(__FILE__));




//=====================================================================
if (!class_exists('Services_JSON')) {

// PEAR JSON class

/**
* Converts to and from JSON format.
*
* JSON (JavaScript Object Notation) is a lightweight data-interchange
* format. It is easy for humans to read and write. It is easy for machines
* to parse and generate. It is based on a subset of the JavaScript
* Programming Language, Standard ECMA-262 3rd Edition - December 1999.
* This feature can also be found in  Python. JSON is a text format that is
* completely language independent but uses conventions that are familiar
* to programmers of the C-family of languages, including C, C++, C#, Java,
* JavaScript, Perl, TCL, and many others. These properties make JSON an
* ideal data-interchange language.
*
* This package provides a simple encoder and decoder for JSON notation. It
* is intended for use with client-side Javascript applications that make
* use of HTTPRequest to perform server communication functions - data can
* be encoded into JSON notation for use in a client-side javascript, or
* decoded from incoming Javascript requests. JSON format is native to
* Javascript, and can be directly eval()'ed with no further parsing
* overhead
*
* All strings should be in ASCII or UTF-8 format!
*
* LICENSE: Redistribution and use in source and binary forms, with or
* without modification, are permitted provided that the following
* conditions are met: Redistributions of source code must retain the
* above copyright notice, this list of conditions and the following
* disclaimer. Redistributions in binary form must reproduce the above
* copyright notice, this list of conditions and the following disclaimer
* in the documentation and/or other materials provided with the
* distribution.
*
* THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
* NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
*
* @category
* @package     Services_JSON
* @author      Michal Migurski <mike-json@teczno.com>
* @author      Matt Knapp <mdknapp[at]gmail[dot]com>
* @author      Brett Stimmerman <brettstimmerman[at]gmail[dot]com>
* @copyright   2005 Michal Migurski
* @version     CVS: $Id: JSON.php,v 1.31 2006/06/28 05:54:17 migurski Exp $
* @license     http://www.opensource.org/licenses/bsd-license.php
* @link        http://pear.php.net/pepr/pepr-proposal-show.php?id=198
*/

/**
* Marker constant for Services_JSON::decode(), used to flag stack state
*/
define('SERVICES_JSON_SLICE',   1);

/**
* Marker constant for Services_JSON::decode(), used to flag stack state
*/
define('SERVICES_JSON_IN_STR',  2);

/**
* Marker constant for Services_JSON::decode(), used to flag stack state
*/
define('SERVICES_JSON_IN_ARR',  3);

/**
* Marker constant for Services_JSON::decode(), used to flag stack state
*/
define('SERVICES_JSON_IN_OBJ',  4);

/**
* Marker constant for Services_JSON::decode(), used to flag stack state
*/
define('SERVICES_JSON_IN_CMT', 5);

/**
* Behavior switch for Services_JSON::decode()
*/
define('SERVICES_JSON_LOOSE_TYPE', 16);

/**
* Behavior switch for Services_JSON::decode()
*/
define('SERVICES_JSON_SUPPRESS_ERRORS', 32);

/**
* Converts to and from JSON format.
*
* Brief example of use:
*
* <code>
* // create a new instance of Services_JSON
* $json = new Services_JSON();
*
* // convert a complexe value to JSON notation, and send it to the browser
* $value = array('foo', 'bar', array(1, 2, 'baz'), array(3, array(4)));
* $output = $json->encode($value);
*
* print($output);
* // prints: ["foo","bar",[1,2,"baz"],[3,[4]]]
*
* // accept incoming POST data, assumed to be in JSON notation
* $input = file_get_contents('php://input', 1000000);
* $value = $json->decode($input);
* </code>
*/
class Services_JSON
{
   /**
    * constructs a new JSON instance
    *
    * @param    int     $use    object behavior flags; combine with boolean-OR
    *
    *                           possible values:
    *                           - SERVICES_JSON_LOOSE_TYPE:  loose typing.
    *                                   "{...}" syntax creates associative arrays
    *                                   instead of objects in decode().
    *                           - SERVICES_JSON_SUPPRESS_ERRORS:  error suppression.
    *                                   Values which can't be encoded (e.g. resources)
    *                                   appear as NULL instead of throwing errors.
    *                                   By default, a deeply-nested resource will
    *                                   bubble up with an error, so all return values
    *                                   from encode() should be checked with isError()
    */
    function Services_JSON($use = 0)
    {
        $this->use = $use;
    }

   /**
    * convert a string from one UTF-16 char to one UTF-8 char
    *
    * Normally should be handled by mb_convert_encoding, but
    * provides a slower PHP-only method for installations
    * that lack the multibye string extension.
    *
    * @param    string  $utf16  UTF-16 character
    * @return   string  UTF-8 character
    * @access   private
    */
    function utf162utf8($utf16)
    {
        // oh please oh please oh please oh please oh please
        if(function_exists('mb_convert_encoding')) {
            return mb_convert_encoding($utf16, 'UTF-8', 'UTF-16');
        }

        $bytes = (ord($utf16{0}) << 8) | ord($utf16{1});

        switch(true) {
            case ((0x7F & $bytes) == $bytes):
                // this case should never be reached, because we are in ASCII range
                // see: http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                return chr(0x7F & $bytes);

            case (0x07FF & $bytes) == $bytes:
                // return a 2-byte UTF-8 character
                // see: http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                return chr(0xC0 | (($bytes >> 6) & 0x1F))
                     . chr(0x80 | ($bytes & 0x3F));

            case (0xFFFF & $bytes) == $bytes:
                // return a 3-byte UTF-8 character
                // see: http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                return chr(0xE0 | (($bytes >> 12) & 0x0F))
                     . chr(0x80 | (($bytes >> 6) & 0x3F))
                     . chr(0x80 | ($bytes & 0x3F));
        }

        // ignoring UTF-32 for now, sorry
        return '';
    }

   /**
    * convert a string from one UTF-8 char to one UTF-16 char
    *
    * Normally should be handled by mb_convert_encoding, but
    * provides a slower PHP-only method for installations
    * that lack the multibye string extension.
    *
    * @param    string  $utf8   UTF-8 character
    * @return   string  UTF-16 character
    * @access   private
    */
    function utf82utf16($utf8)
    {
        // oh please oh please oh please oh please oh please
        if(function_exists('mb_convert_encoding')) {
            return mb_convert_encoding($utf8, 'UTF-16', 'UTF-8');
        }

        switch(strlen($utf8)) {
            case 1:
                // this case should never be reached, because we are in ASCII range
                // see: http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                return $utf8;

            case 2:
                // return a UTF-16 character from a 2-byte UTF-8 char
                // see: http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                return chr(0x07 & (ord($utf8{0}) >> 2))
                     . chr((0xC0 & (ord($utf8{0}) << 6))
                         | (0x3F & ord($utf8{1})));

            case 3:
                // return a UTF-16 character from a 3-byte UTF-8 char
                // see: http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                return chr((0xF0 & (ord($utf8{0}) << 4))
                         | (0x0F & (ord($utf8{1}) >> 2)))
                     . chr((0xC0 & (ord($utf8{1}) << 6))
                         | (0x7F & ord($utf8{2})));
        }

        // ignoring UTF-32 for now, sorry
        return '';
    }

   /**
    * encodes an arbitrary variable into JSON format
    *
    * @param    mixed   $var    any number, boolean, string, array, or object to be encoded.
    *                           see argument 1 to Services_JSON() above for array-parsing behavior.
    *                           if var is a strng, note that encode() always expects it
    *                           to be in ASCII or UTF-8 format!
    *
    * @return   mixed   JSON string representation of input var or an error if a problem occurs
    * @access   public
    */
    function encode($var)
    {
        switch (gettype($var)) {
            case 'boolean':
                return $var ? 'true' : 'false';

            case 'NULL':
                return 'null';

            case 'integer':
                return (int) $var;

            case 'double':
            case 'float':
                return (float) $var;

            case 'string':
                // STRINGS ARE EXPECTED TO BE IN ASCII OR UTF-8 FORMAT
                $ascii = '';
                $strlen_var = strlen($var);

               /*
                * Iterate over every character in the string,
                * escaping with a slash or encoding to UTF-8 where necessary
                */
                for ($c = 0; $c < $strlen_var; ++$c) {

                    $ord_var_c = ord($var{$c});

                    switch (true) {
                        case $ord_var_c == 0x08:
                            $ascii .= '\b';
                            break;
                        case $ord_var_c == 0x09:
                            $ascii .= '\t';
                            break;
                        case $ord_var_c == 0x0A:
                            $ascii .= '\n';
                            break;
                        case $ord_var_c == 0x0C:
                            $ascii .= '\f';
                            break;
                        case $ord_var_c == 0x0D:
                            $ascii .= '\r';
                            break;

                        case $ord_var_c == 0x22:
                        case $ord_var_c == 0x2F:
                        case $ord_var_c == 0x5C:
                            // double quote, slash, slosh
                            $ascii .= '\\'.$var{$c};
                            break;

                        case (($ord_var_c >= 0x20) && ($ord_var_c <= 0x7F)):
                            // characters U-00000000 - U-0000007F (same as ASCII)
                            $ascii .= $var{$c};
                            break;

                        case (($ord_var_c & 0xE0) == 0xC0):
                            // characters U-00000080 - U-000007FF, mask 110XXXXX
                            // see http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                            $char = pack('C*', $ord_var_c, ord($var{$c + 1}));
                            $c += 1;
                            $utf16 = $this->utf82utf16($char);
                            $ascii .= sprintf('\u%04s', bin2hex($utf16));
                            break;

                        case (($ord_var_c & 0xF0) == 0xE0):
                            // characters U-00000800 - U-0000FFFF, mask 1110XXXX
                            // see http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                            $char = pack('C*', $ord_var_c,
                                         ord($var{$c + 1}),
                                         ord($var{$c + 2}));
                            $c += 2;
                            $utf16 = $this->utf82utf16($char);
                            $ascii .= sprintf('\u%04s', bin2hex($utf16));
                            break;

                        case (($ord_var_c & 0xF8) == 0xF0):
                            // characters U-00010000 - U-001FFFFF, mask 11110XXX
                            // see http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                            $char = pack('C*', $ord_var_c,
                                         ord($var{$c + 1}),
                                         ord($var{$c + 2}),
                                         ord($var{$c + 3}));
                            $c += 3;
                            $utf16 = $this->utf82utf16($char);
                            $ascii .= sprintf('\u%04s', bin2hex($utf16));
                            break;

                        case (($ord_var_c & 0xFC) == 0xF8):
                            // characters U-00200000 - U-03FFFFFF, mask 111110XX
                            // see http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                            $char = pack('C*', $ord_var_c,
                                         ord($var{$c + 1}),
                                         ord($var{$c + 2}),
                                         ord($var{$c + 3}),
                                         ord($var{$c + 4}));
                            $c += 4;
                            $utf16 = $this->utf82utf16($char);
                            $ascii .= sprintf('\u%04s', bin2hex($utf16));
                            break;

                        case (($ord_var_c & 0xFE) == 0xFC):
                            // characters U-04000000 - U-7FFFFFFF, mask 1111110X
                            // see http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                            $char = pack('C*', $ord_var_c,
                                         ord($var{$c + 1}),
                                         ord($var{$c + 2}),
                                         ord($var{$c + 3}),
                                         ord($var{$c + 4}),
                                         ord($var{$c + 5}));
                            $c += 5;
                            $utf16 = $this->utf82utf16($char);
                            $ascii .= sprintf('\u%04s', bin2hex($utf16));
                            break;
                    }
                }

                return '"'.$ascii.'"';

            case 'array':
               /*
                * As per JSON spec if any array key is not an integer
                * we must treat the the whole array as an object. We
                * also try to catch a sparsely populated associative
                * array with numeric keys here because some JS engines
                * will create an array with empty indexes up to
                * max_index which can cause memory issues and because
                * the keys, which may be relevant, will be remapped
                * otherwise.
                *
                * As per the ECMA and JSON specification an object may
                * have any string as a property. Unfortunately due to
                * a hole in the ECMA specification if the key is a
                * ECMA reserved word or starts with a digit the
                * parameter is only accessible using ECMAScript's
                * bracket notation.
                */

                // treat as a JSON object
                if (is_array($var) && count($var) && (array_keys($var) !== range(0, sizeof($var) - 1))) {
                    $properties = array_map(array($this, 'name_value'),
                                            array_keys($var),
                                            array_values($var));

                    foreach($properties as $property) {
                        if(Services_JSON::isError($property)) {
                            return $property;
                        }
                    }

                    return '{' . join(',', $properties) . '}';
                }

                // treat it like a regular array
                $elements = array_map(array($this, 'encode'), $var);

                foreach($elements as $element) {
                    if(Services_JSON::isError($element)) {
                        return $element;
                    }
                }

                return '[' . join(',', $elements) . ']';

            case 'object':
                $vars = get_object_vars($var);

                $properties = array_map(array($this, 'name_value'),
                                        array_keys($vars),
                                        array_values($vars));

                foreach($properties as $property) {
                    if(Services_JSON::isError($property)) {
                        return $property;
                    }
                }

                return '{' . join(',', $properties) . '}';

            default:
                return ($this->use & SERVICES_JSON_SUPPRESS_ERRORS)
                    ? 'null'
                    : new Services_JSON_Error(gettype($var)." can not be encoded as JSON string");
        }
    }

   /**
    * array-walking function for use in generating JSON-formatted name-value pairs
    *
    * @param    string  $name   name of key to use
    * @param    mixed   $value  reference to an array element to be encoded
    *
    * @return   string  JSON-formatted name-value pair, like '"name":value'
    * @access   private
    */
    function name_value($name, $value)
    {
        $encoded_value = $this->encode($value);

        if(Services_JSON::isError($encoded_value)) {
            return $encoded_value;
        }

        return $this->encode(strval($name)) . ':' . $encoded_value;
    }

   /**
    * reduce a string by removing leading and trailing comments and whitespace
    *
    * @param    $str    string      string value to strip of comments and whitespace
    *
    * @return   string  string value stripped of comments and whitespace
    * @access   private
    */
    function reduce_string($str)
    {
        $str = preg_replace(array(

                // eliminate single line comments in '// ...' form
                '#^\s*//(.+)$#m',

                // eliminate multi-line comments in '/* ... */' form, at start of string
                '#^\s*/\*(.+)\*/#Us',

                // eliminate multi-line comments in '/* ... */' form, at end of string
                '#/\*(.+)\*/\s*$#Us'

            ), '', $str);

        // eliminate extraneous space
        return trim($str);
    }

   /**
    * decodes a JSON string into appropriate variable
    *
    * @param    string  $str    JSON-formatted string
    *
    * @return   mixed   number, boolean, string, array, or object
    *                   corresponding to given JSON input string.
    *                   See argument 1 to Services_JSON() above for object-output behavior.
    *                   Note that decode() always returns strings
    *                   in ASCII or UTF-8 format!
    * @access   public
    */
    function decode($str)
    {
        $str = $this->reduce_string($str);

        switch (strtolower($str)) {
            case 'true':
                return true;

            case 'false':
                return false;

            case 'null':
                return null;

            default:
                $m = array();

                if (is_numeric($str)) {
                    // Lookie-loo, it's a number

                    // This would work on its own, but I'm trying to be
                    // good about returning integers where appropriate:
                    // return (float)$str;

                    // Return float or int, as appropriate
                    return ((float)$str == (integer)$str)
                        ? (integer)$str
                        : (float)$str;

                } elseif (preg_match('/^("|\').*(\1)$/s', $str, $m) && $m[1] == $m[2]) {
                    // STRINGS RETURNED IN UTF-8 FORMAT
                    $delim = substr($str, 0, 1);
                    $chrs = substr($str, 1, -1);
                    $utf8 = '';
                    $strlen_chrs = strlen($chrs);

                    for ($c = 0; $c < $strlen_chrs; ++$c) {

                        $substr_chrs_c_2 = substr($chrs, $c, 2);
                        $ord_chrs_c = ord($chrs{$c});

                        switch (true) {
                            case $substr_chrs_c_2 == '\b':
                                $utf8 .= chr(0x08);
                                ++$c;
                                break;
                            case $substr_chrs_c_2 == '\t':
                                $utf8 .= chr(0x09);
                                ++$c;
                                break;
                            case $substr_chrs_c_2 == '\n':
                                $utf8 .= chr(0x0A);
                                ++$c;
                                break;
                            case $substr_chrs_c_2 == '\f':
                                $utf8 .= chr(0x0C);
                                ++$c;
                                break;
                            case $substr_chrs_c_2 == '\r':
                                $utf8 .= chr(0x0D);
                                ++$c;
                                break;

                            case $substr_chrs_c_2 == '\\"':
                            case $substr_chrs_c_2 == '\\\'':
                            case $substr_chrs_c_2 == '\\\\':
                            case $substr_chrs_c_2 == '\\/':
                                if (($delim == '"' && $substr_chrs_c_2 != '\\\'') ||
                                   ($delim == "'" && $substr_chrs_c_2 != '\\"')) {
                                    $utf8 .= $chrs{++$c};
                                }
                                break;

                            case preg_match('/\\\u[0-9A-F]{4}/i', substr($chrs, $c, 6)):
                                // single, escaped unicode character
                                $utf16 = chr(hexdec(substr($chrs, ($c + 2), 2)))
                                       . chr(hexdec(substr($chrs, ($c + 4), 2)));
                                $utf8 .= $this->utf162utf8($utf16);
                                $c += 5;
                                break;

                            case ($ord_chrs_c >= 0x20) && ($ord_chrs_c <= 0x7F):
                                $utf8 .= $chrs{$c};
                                break;

                            case ($ord_chrs_c & 0xE0) == 0xC0:
                                // characters U-00000080 - U-000007FF, mask 110XXXXX
                                //see http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                                $utf8 .= substr($chrs, $c, 2);
                                ++$c;
                                break;

                            case ($ord_chrs_c & 0xF0) == 0xE0:
                                // characters U-00000800 - U-0000FFFF, mask 1110XXXX
                                // see http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                                $utf8 .= substr($chrs, $c, 3);
                                $c += 2;
                                break;

                            case ($ord_chrs_c & 0xF8) == 0xF0:
                                // characters U-00010000 - U-001FFFFF, mask 11110XXX
                                // see http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                                $utf8 .= substr($chrs, $c, 4);
                                $c += 3;
                                break;

                            case ($ord_chrs_c & 0xFC) == 0xF8:
                                // characters U-00200000 - U-03FFFFFF, mask 111110XX
                                // see http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                                $utf8 .= substr($chrs, $c, 5);
                                $c += 4;
                                break;

                            case ($ord_chrs_c & 0xFE) == 0xFC:
                                // characters U-04000000 - U-7FFFFFFF, mask 1111110X
                                // see http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
                                $utf8 .= substr($chrs, $c, 6);
                                $c += 5;
                                break;

                        }

                    }

                    return $utf8;

                } elseif (preg_match('/^\[.*\]$/s', $str) || preg_match('/^\{.*\}$/s', $str)) {
                    // array, or object notation

                    if ($str{0} == '[') {
                        $stk = array(SERVICES_JSON_IN_ARR);
                        $arr = array();
                    } else {
                        if ($this->use & SERVICES_JSON_LOOSE_TYPE) {
                            $stk = array(SERVICES_JSON_IN_OBJ);
                            $obj = array();
                        } else {
                            $stk = array(SERVICES_JSON_IN_OBJ);
                            $obj = new stdClass();
                        }
                    }

                    array_push($stk, array('what'  => SERVICES_JSON_SLICE,
                                           'where' => 0,
                                           'delim' => false));

                    $chrs = substr($str, 1, -1);
                    $chrs = $this->reduce_string($chrs);

                    if ($chrs == '') {
                        if (reset($stk) == SERVICES_JSON_IN_ARR) {
                            return $arr;

                        } else {
                            return $obj;

                        }
                    }

                    //print("\nparsing {$chrs}\n");

                    $strlen_chrs = strlen($chrs);

                    for ($c = 0; $c <= $strlen_chrs; ++$c) {

                        $top = end($stk);
                        $substr_chrs_c_2 = substr($chrs, $c, 2);

                        if (($c == $strlen_chrs) || (($chrs{$c} == ',') && ($top['what'] == SERVICES_JSON_SLICE))) {
                            // found a comma that is not inside a string, array, etc.,
                            // OR we've reached the end of the character list
                            $slice = substr($chrs, $top['where'], ($c - $top['where']));
                            array_push($stk, array('what' => SERVICES_JSON_SLICE, 'where' => ($c + 1), 'delim' => false));
                            //print("Found split at {$c}: ".substr($chrs, $top['where'], (1 + $c - $top['where']))."\n");

                            if (reset($stk) == SERVICES_JSON_IN_ARR) {
                                // we are in an array, so just push an element onto the stack
                                array_push($arr, $this->decode($slice));

                            } elseif (reset($stk) == SERVICES_JSON_IN_OBJ) {
                                // we are in an object, so figure
                                // out the property name and set an
                                // element in an associative array,
                                // for now
                                $parts = array();
                                
                                if (preg_match('/^\s*(["\'].*[^\\\]["\'])\s*:\s*(\S.*),?$/Uis', $slice, $parts)) {
                                    // "name":value pair
                                    $key = $this->decode($parts[1]);
                                    $val = $this->decode($parts[2]);

                                    if ($this->use & SERVICES_JSON_LOOSE_TYPE) {
                                        $obj[$key] = $val;
                                    } else {
                                        $obj->$key = $val;
                                    }
                                } elseif (preg_match('/^\s*(\w+)\s*:\s*(\S.*),?$/Uis', $slice, $parts)) {
                                    // name:value pair, where name is unquoted
                                    $key = $parts[1];
                                    $val = $this->decode($parts[2]);

                                    if ($this->use & SERVICES_JSON_LOOSE_TYPE) {
                                        $obj[$key] = $val;
                                    } else {
                                        $obj->$key = $val;
                                    }
                                }

                            }

                        } elseif ((($chrs{$c} == '"') || ($chrs{$c} == "'")) && ($top['what'] != SERVICES_JSON_IN_STR)) {
                            // found a quote, and we are not inside a string
                            array_push($stk, array('what' => SERVICES_JSON_IN_STR, 'where' => $c, 'delim' => $chrs{$c}));
                            //print("Found start of string at {$c}\n");

                        } elseif (($chrs{$c} == $top['delim']) &&
                                 ($top['what'] == SERVICES_JSON_IN_STR) &&
                                 ((strlen(substr($chrs, 0, $c)) - strlen(rtrim(substr($chrs, 0, $c), '\\'))) % 2 != 1)) {
                            // found a quote, we're in a string, and it's not escaped
                            // we know that it's not escaped becase there is _not_ an
                            // odd number of backslashes at the end of the string so far
                            array_pop($stk);
                            //print("Found end of string at {$c}: ".substr($chrs, $top['where'], (1 + 1 + $c - $top['where']))."\n");

                        } elseif (($chrs{$c} == '[') &&
                                 in_array($top['what'], array(SERVICES_JSON_SLICE, SERVICES_JSON_IN_ARR, SERVICES_JSON_IN_OBJ))) {
                            // found a left-bracket, and we are in an array, object, or slice
                            array_push($stk, array('what' => SERVICES_JSON_IN_ARR, 'where' => $c, 'delim' => false));
                            //print("Found start of array at {$c}\n");

                        } elseif (($chrs{$c} == ']') && ($top['what'] == SERVICES_JSON_IN_ARR)) {
                            // found a right-bracket, and we're in an array
                            array_pop($stk);
                            //print("Found end of array at {$c}: ".substr($chrs, $top['where'], (1 + $c - $top['where']))."\n");

                        } elseif (($chrs{$c} == '{') &&
                                 in_array($top['what'], array(SERVICES_JSON_SLICE, SERVICES_JSON_IN_ARR, SERVICES_JSON_IN_OBJ))) {
                            // found a left-brace, and we are in an array, object, or slice
                            array_push($stk, array('what' => SERVICES_JSON_IN_OBJ, 'where' => $c, 'delim' => false));
                            //print("Found start of object at {$c}\n");

                        } elseif (($chrs{$c} == '}') && ($top['what'] == SERVICES_JSON_IN_OBJ)) {
                            // found a right-brace, and we're in an object
                            array_pop($stk);
                            //print("Found end of object at {$c}: ".substr($chrs, $top['where'], (1 + $c - $top['where']))."\n");

                        } elseif (($substr_chrs_c_2 == '/*') &&
                                 in_array($top['what'], array(SERVICES_JSON_SLICE, SERVICES_JSON_IN_ARR, SERVICES_JSON_IN_OBJ))) {
                            // found a comment start, and we are in an array, object, or slice
                            array_push($stk, array('what' => SERVICES_JSON_IN_CMT, 'where' => $c, 'delim' => false));
                            $c++;
                            //print("Found start of comment at {$c}\n");

                        } elseif (($substr_chrs_c_2 == '*/') && ($top['what'] == SERVICES_JSON_IN_CMT)) {
                            // found a comment end, and we're in one now
                            array_pop($stk);
                            $c++;

                            for ($i = $top['where']; $i <= $c; ++$i)
                                $chrs = substr_replace($chrs, ' ', $i, 1);

                            //print("Found end of comment at {$c}: ".substr($chrs, $top['where'], (1 + $c - $top['where']))."\n");

                        }

                    }

                    if (reset($stk) == SERVICES_JSON_IN_ARR) {
                        return $arr;

                    } elseif (reset($stk) == SERVICES_JSON_IN_OBJ) {
                        return $obj;

                    }

                }
        }
    }

    /**
     * @todo Ultimately, this should just call PEAR::isError()
     */
    function isError($data, $code = null)
    {
        if (class_exists('pear')) {
            return PEAR::isError($data, $code);
        } elseif (is_object($data) && (get_class($data) == 'services_json_error' ||
                                 is_subclass_of($data, 'services_json_error'))) {
            return true;
        }

        return false;
    }
}

if (class_exists('PEAR_Error')) {

    class Services_JSON_Error extends PEAR_Error
    {
        function Services_JSON_Error($message = 'unknown error', $code = null,
                                     $mode = null, $options = null, $userinfo = null)
        {
            parent::PEAR_Error($message, $code, $mode, $options, $userinfo);
        }
    }

} else {

    /**
     * @todo Ultimately, this class shall be descended from PEAR_Error
     */
    class Services_JSON_Error
    {
        function Services_JSON_Error($message = 'unknown error', $code = null,
                                     $mode = null, $options = null, $userinfo = null)
        {

        }
    }

}


}


?>
